# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import datetime as dt
from typing import Sequence, overload

from lenskit.data import Dataset, DatasetBuilder, ItemListCollection
from lenskit.logging import get_logger

from .split import TTSplit

_log = get_logger(__name__)


@overload
def split_global_time(
    data: Dataset,
    time: int | float | str | dt.datetime,
    end: int | float | str | dt.datetime | None = None,
) -> TTSplit: ...
@overload
def split_global_time(
    data: Dataset,
    time: Sequence[int | float | str | dt.datetime],
    end: int | float | str | dt.datetime | None = None,
) -> list[TTSplit]: ...
def split_global_time(
    data: Dataset,
    time: int | float | str | dt.datetime | Sequence[int | float | str | dt.datetime],
    end: int | float | str | dt.datetime | None = None,
) -> TTSplit | list[TTSplit]:
    """
    Global temporal train-test split.  This splits a data set into train/test
    pairs using a single global timestamp.  When given multiple timestamps, it
    will return multiple splits, where split :math:`i` has training data from
    before :math:`t_i` and testing data on or after :math:`t_i` and before
    :math:`t_{i+1}` (the last split has no upper bound on the testing data).

    Stability:
        Caller

    Args:
        data:
            The dataset to split.
        time:
            Time or sequence of times at which to split.  Strings must be in ISO
            format.
        end:
            A final cutoff time for the testing data.

    Returns:
        The data splits.
    """
    log = _log.bind(n_records=data.interaction_count)
    if isinstance(time, (str, int, float, dt.datetime)):
        times = [_make_time(time)]
        rv = "single"
    else:
        times = [_make_time(t) for t in time]
        rv = "sequence"
        log = log.bind(n_splits=len(times))

    iname = data.default_interaction_class()
    matrix = data.interactions().pandas(ids=True)
    if "timestamp" not in matrix:
        raise RuntimeError("temporal split requires timestamp")

    ts_col = matrix["timestamp"]
    # ts_col = np.asarray(ts_col)

    if ts_col.dtype.kind in ("i", "u", "f"):
        log.debug("converting query timestamps")
        times = [t.timestamp() for t in times]

    results = []
    for i, t in enumerate(times):
        tlog = log.bind(number=i, test_start=t)
        tlog.debug("creating initial split")
        mask = ts_col >= t
        train_build = DatasetBuilder(data)
        train_build.filter_interactions(iname, max_time=t)

        t2 = end
        if i + 1 < len(times):
            t2 = times[i + 1]

        if t2 is None:
            test = matrix[mask]
        else:
            tlog = tlog.bind(test_end=t2)
            tlog.debug("filtering test data for upper bound")
            test = matrix[mask & (ts_col < t2)]

        tlog.debug("building training data set")
        train_ds = train_build.build()
        tlog.debug("building testing item lists")
        test_ilc = ItemListCollection.from_df(test, ["user_id"])
        tlog.debug("built split with %d train interactions", train_ds.interaction_count)
        results.append(TTSplit(train_ds, test_ilc))

    if rv == "sequence":
        return results
    else:
        assert len(results) == 1
        return results[0]


def split_temporal_fraction(data: Dataset, test_fraction: float) -> TTSplit:
    """
    Do a global temporal split of a data set based on a test set size.

    Args:
        data:
            The dataset to split.
        test_fraction:
            The fraction of the interactions to put in the testing data.
    """
    df = data.interaction_table(format="pandas")
    point = df["timestamp"].quantile(1 - test_fraction)
    return split_global_time(data, point)


def _make_time(t: int | float | str | dt.datetime) -> dt.datetime:
    if isinstance(t, (int, float)):
        return dt.datetime.fromtimestamp(t)
    elif isinstance(t, str):
        return dt.datetime.fromisoformat(t)
    else:
        return t



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import numpy as np
import pandas as pd
from typing_extensions import Literal, override

from lenskit.data.dataset import Dataset
from lenskit.data.items import ItemList

from ._base import ListMetric, RankingMetricBase


class MeanPopRank(ListMetric, RankingMetricBase):
    r"""
    Compute the _obscurity_ (mean popularity rank) of the recommendations.

    Unlike other metrics, this metric requires access to the training dataset in
    order to compute item popularity metrics.  Supply this as a constructor
    parameter.

    This metric represents the popularity rank as a quantile, based on the
    either the number of distinct users who have interacted with the item, or
    the total interactions (depending on the options — distinct users is the
    default).

    Let $q_i$ be the _popularity rank_, represented as a quantile, of item $i$.
    $q_i = 1$ for the most-popular item; $q_i=0$ for an item with no users or
    interactions (the quantiles are min-max scaled). This metric computes the
    mean of the quantile popularity ranks for the recommended items:

    .. math::
        \mathcal{M}(L) = \frac{1}{|L|} \sum_{i \in L} q_i

    This metric is based on the ``obscurity'' metric of
    :cite:t:`ekstrandSturgeonCoolKids2017` and the popularity-based item novelty
    metric of :cite:t:`vargasRankRelevanceNovelty2011`.

    Stability:
        Caller
    """

    item_ranks: pd.Series[float]

    def __init__(
        self, data: Dataset, k: int | None = None, count: Literal["users", "interactions"] = "users"
    ):
        super().__init__(k=k)
        stats = data.item_stats()
        match count:
            case "users":
                counts = stats["user_count"]
            case "interactions":
                counts = stats["count"]
            case _:  # pragma: nocover
                raise ValueError(f"invalid count {count}")

        # If we just computed quantiles, items with count 0 would have a small positive
        # count. Let's quantile the positive items, then add back in zeros.
        pos = counts[counts > 0]
        ranks = pos.rank(method="average", ascending=True)
        ranks /= len(pos)
        self.item_ranks = ranks.reindex(counts.index, fill_value=0)

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        recs = self.truncate(recs)
        nrecs = len(recs)
        if nrecs == 0:
            return np.nan

        items = recs.ids()
        ranks = self.item_ranks.reindex(items, fill_value=0)
        return ranks.mean()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Data attribute accessors.
"""

from __future__ import annotations

import json
from typing import Literal

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import torch
from numpy.typing import NDArray
from scipy.sparse import csr_array
from typing_extensions import Any

from .schema import AttrLayout, ColumnSpec
from .types import IDArray
from .vocab import Vocabulary


def attr_set(
    name: str, spec: ColumnSpec, table: pa.Table, vocab: Vocabulary, rows: pa.Int32Array | None
):
    match spec.layout:
        case AttrLayout.SCALAR:
            return ScalarAttributeSet(name, spec, table, vocab, rows)
        case AttrLayout.LIST:
            return ListAttributeSet(name, spec, table, vocab, rows)
        case AttrLayout.VECTOR:
            return VectorAttributeSet(name, spec, table, vocab, rows)
        case AttrLayout.SPARSE:
            return SparseAttributeSet(name, spec, table, vocab, rows)
        case _:  # pragma: nocover
            raise ValueError(f"unsupported layout {spec.layout}")


class AttributeSet:
    """
    Base class for attributes associated with entities.

    This is the general interface for attribute sets.  Not all access methods
    are supported for all layouts.

    Stability:
        caller
    """

    name: str
    """
    The name of the attribute.
    """
    _spec: ColumnSpec
    _table: pa.Table
    _vocab: Vocabulary
    _selected: pa.Int32Array | None = None

    def __init__(
        self,
        name: str,
        spec: ColumnSpec,
        table: pa.Table,
        vocab: Vocabulary,
        rows: pa.Int32Array | None,
    ):
        self.name = name
        self._spec = spec
        self._table = table
        self._vocab = vocab
        self._selected = rows

    def ids(self) -> IDArray:
        """
        Get the entity IDs for the rows in this attribute's values.
        """
        if self._selected is None:
            return self._vocab.ids()
        else:
            return self._vocab.ids(self._selected.to_numpy())

    def numbers(self) -> np.ndarray[int, np.dtype[np.int32]]:
        """
        Get the entity numbers for the rows in this attribute's values.
        """
        if self._selected is None:
            return np.arange(self._table.num_rows, dtype=np.int32)
        else:
            return self._selected.to_numpy()

    @property
    def dim_names(self) -> list[str] | None:
        """
        Get the names attached to this attribute's dimensions.

        .. note::

            Only applicable to vector and sparse attributes.
        """
        return None

    @property
    def is_scalar(self) -> bool:
        """
        Query whether this attribute is scalar.
        """
        return self._spec.layout == AttrLayout.SCALAR

    @property
    def is_list(self) -> bool:
        """
        Query whether this attribute is a list.
        """
        return self._spec.layout == AttrLayout.LIST

    @property
    def is_vector(self) -> bool:
        """
        Query whether this attribute is a dense vector.
        """
        return self._spec.layout == AttrLayout.VECTOR

    @property
    def is_sparse(self) -> bool:
        """
        Query whether this attribute is a sparse vector.
        """
        return self._spec.layout == AttrLayout.SPARSE

    def pandas(
        self, *, missing: Literal["null", "omit"] = "null"
    ) -> pd.Series | pd.DataFrame:  # pragma: nocover
        raise NotImplementedError()

    def numpy(self) -> NDArray[Any]:
        """
        Get the attribute values as a NumPy array.

        .. note::
            Undefined attribute values may have undefined contents; they will
            _usually_ be ``NaN`` or similar, but this is not fully guaranteed.
        """
        return self.arrow().to_numpy()

    def arrow(self) -> pa.Array[Any] | pa.ChunkedArray[Any]:
        """
        Get the attribute values as an Arrow array.
        """
        col = self._table.column(self.name)
        if self._selected is not None:
            col = col.take(self._selected)

        return col

    def scipy(self) -> NDArray[Any] | csr_array:
        """
        Get this attribute as a SciPy sparse array (if it is sparse), or a NumPy
        array if it is dense.
        """
        return self.numpy()

    def torch(self) -> torch.Tensor:
        return torch.from_numpy(self.numpy())

    def drop_null(self):
        """
        Subset this attribute set to only entities for which it is defined.
        """
        col = self._table.column(self.name)
        valid = col.is_valid().combine_chunks()
        if self._selected is not None:
            valid = valid.take(self._selected)
            selected = self._selected.filter(valid)
        else:
            selected = np.arange(len(self._vocab), dtype=np.int32)
            selected = pa.array(selected)
            selected = selected.filter(valid)

        return self.__class__(self.name, self._spec, self._table, self._vocab, selected)

    def __len__(self):
        if self._selected is not None:
            return len(self._selected)
        else:
            return len(self._vocab)


class ScalarAttributeSet(AttributeSet):
    def pandas(self, *, missing: Literal["null", "omit"] = "null") -> pd.Series[Any]:
        arr = self.arrow()
        mask = arr.is_valid()
        if missing == "null" and pc.all(mask).as_py():
            return pd.Series(arr.to_numpy(zero_copy_only=False), index=self.ids())
        else:
            mask = mask.to_numpy(zero_copy_only=False)
            return pd.Series(arr.drop_null().to_numpy(zero_copy_only=False), index=self.ids()[mask])


class ListAttributeSet(AttributeSet):
    def pandas(self, *, missing: Literal["null", "omit"] = "null") -> pd.Series[Any]:
        arr = self.arrow()
        mask = arr.is_valid()
        if missing == "null" and pc.all(mask).as_py():
            return pd.Series(arr.to_numpy(zero_copy_only=False), index=self.ids())
        else:
            mask = mask.to_numpy(zero_copy_only=False)
            return pd.Series(arr.drop_null().to_numpy(zero_copy_only=False), index=self.ids()[mask])


class VectorAttributeSet(AttributeSet):
    _names: list[str] | None = None
    _size: int | None = None

    @property
    def dim_names(self) -> list[str] | None:
        if self._names is None:
            field = self._table.field(self.name)
            meta = field.metadata
            if meta is None:
                return None

            nstr = meta.get(b"lenskit:names", None) if meta else None
            if nstr is not None:
                self._names = json.loads(nstr)

        return self._names

    @property
    def vector_size(self) -> int:
        assert self._spec.vector_size is not None, "vector column has no size"
        return self._spec.vector_size

    def arrow(self) -> pa.Array[Any] | pa.ChunkedArray[Any]:
        """
        Get the attribute values as an Arrow array.
        """

        col = self._table.column(self.name)
        if self._selected is not None:
            col = col.take(self._selected)

        if pa.types.is_fixed_size_list(col.type):
            return col
        elif pc.all(col.is_valid()).as_py():
            if isinstance(col, pa.ChunkedArray):
                return pa.chunked_array(
                    [
                        pa.FixedSizeListArray.from_arrays(c.values, self.vector_size)
                        for c in col.chunks
                    ]
                )
            elif isinstance(col, pa.ListArray):
                return pa.FixedSizeListArray.from_arrays(col.values, self.vector_size)
            else:  # pragma: nocover
                raise TypeError("unexpected array type")
        else:
            # now things get tricky.
            if isinstance(col, pa.ChunkedArray):
                col = col.combine_chunks()
            assert isinstance(col, pa.ListArray)
            fixed = pa.FixedSizeListArray.from_arrays(col.values, self.vector_size)
            mat = pa.nulls(len(col), type=fixed.type)
            valid = col.is_valid()
            assert pc.all(pc.equal(col.value_lengths().filter(valid), self.vector_size)).as_py()
            col = _replace_vectors(mat, valid, fixed)

        return col

    def numpy(self) -> np.ndarray[tuple[int, int], Any]:
        arr = self.arrow()
        if isinstance(arr, pa.ChunkedArray):
            arr = arr.combine_chunks()
        assert isinstance(arr, pa.FixedSizeListArray)
        mat = arr.values.to_numpy(zero_copy_only=False).reshape((len(arr), arr.type.list_size))
        return mat

    def pandas(self, *, missing: Literal["null", "omit"] = "null") -> pd.DataFrame:
        arr = self.arrow()
        if isinstance(arr, pa.ChunkedArray):
            arr = arr.combine_chunks()
        assert isinstance(arr, pa.FixedSizeListArray)

        ids = self.ids()

        mask = arr.is_valid()
        if missing == "omit" and not pc.all(mask).as_py():
            arr = arr.drop_null()
            ids = ids[mask.to_numpy(zero_copy_only=False)]

        mat = arr.values.to_numpy().reshape((len(arr), arr.type.list_size))
        return pd.DataFrame(mat, index=ids, columns=self.dim_names)


class SparseAttributeSet(AttributeSet):
    _names: list[str] | None = None

    @property
    def dim_names(self) -> list[str] | None:
        if self._names is None:
            field = self._table.field(self.name)
            meta = field.metadata
            if meta is None:
                return None

            nstr = meta.get(b"lenskit:names", None) if meta else None
            if nstr is not None:
                self._names = json.loads(nstr)

        return self._names

    def numpy(self) -> np.ndarray[tuple[int, int], Any]:
        raise NotImplementedError("sparse attributes cannot be retrieved as numpy")

    def scipy(self) -> csr_array:
        col = self.arrow()
        if isinstance(col, pa.ChunkedArray):
            col = col.combine_chunks()
        assert isinstance(col, pa.ListArray)

        rowptr = col.offsets.to_numpy()
        entries = col.values
        assert isinstance(entries, pa.StructArray)
        indices = entries.field("index").to_numpy(zero_copy_only=False)
        values = entries.field("value").to_numpy(zero_copy_only=False)

        ncol = self._spec.vector_size
        assert ncol is not None, "sparse vector has no size"

        return csr_array((values, indices, rowptr), shape=(len(col), ncol))

    def torch(self) -> torch.Tensor:
        csr = self.scipy()
        return torch.sparse_csr_tensor(
            crow_indices=np.require(csr.indptr, requirements="W"),
            col_indices=np.require(csr.indices, requirements="W"),
            values=np.require(csr.data, requirements="W"),
            size=csr.shape,
        )


def _replace_vectors(
    arr: pa.FixedSizeListArray, mask: pa.BooleanArray, values: pa.FixedSizeListArray
):
    size = arr.type.list_size
    assert values.type.list_size == size
    assert pc.all(values.is_valid()).as_py()
    in_valid = arr.is_valid()
    out_valid = pc.or_(in_valid, mask)

    value_mask = np.repeat(np.asarray(mask), size)
    value_mask = pa.array(value_mask)
    new_vals = pc.replace_with_mask(arr.values, value_mask, values.values)
    return pa.FixedSizeListArray.from_arrays(new_vals, size, mask=pc.invert(out_valid))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
from typing_extensions import override

from lenskit.data import ItemList

from ._base import ListMetric, RankingMetricBase


class Hit(ListMetric, RankingMetricBase):
    """
    Compute whether or not a list is a hit; any list with at least one
    relevant item in the first :math:`k` positions (:math:`L_{\\le k} \\cap
    I_u^{\\mathrm{test}} \\ne \\emptyset`) is scored as 1, and lists with no
    relevant items as 0.  When averaged over the recommendation lists, this
    computes the *hit rate* :cite:p:`deshpande:iknn`.

    Stability:
        Caller
    """

    @property
    def label(self):
        if self.k is not None:
            return f"Hit@{self.k}"
        else:
            return "Hit"

    @override
    def measure_list(self, recs: ItemList, test: ItemList) -> float:
        if len(test) == 0:
            return np.nan

        recs = self.truncate(recs)

        return 1 if np.any(np.isin(recs.ids(), test.ids())) else 0



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Data abstractions and data set access.
"""

from __future__ import annotations

from lenskit.diagnostics import FieldError

from .adapt import from_interactions_df
from .attributes import AttributeSet
from .builder import DatasetBuilder
from .collection import (
    GenericKey,
    ItemListCollection,
    ItemListCollector,
    ListILC,
    MutableItemListCollection,
    UserIDKey,
)
from .dataset import Dataset, EntitySet, MatrixRelationshipSet, RelationshipSet
from .items import ItemList
from .matrix import COOStructure, CSRStructure
from .movielens import load_movielens, load_movielens_df
from .mtarray import MTArray, MTFloatArray, MTGenericArray, MTIntArray
from .query import QueryInput, RecQuery
from .types import ID, NPID, FeedbackType
from .vocab import Vocabulary

__all__ = [
    "Dataset",
    "EntitySet",
    "RelationshipSet",
    "MatrixRelationshipSet",
    "AttributeSet",
    "DatasetBuilder",
    "FieldError",
    "CSRStructure",
    "COOStructure",
    "from_interactions_df",
    "ID",
    "NPID",
    "FeedbackType",
    "ItemList",
    "ItemListCollection",
    "ItemListCollector",
    "MutableItemListCollection",
    "ListILC",
    "UserIDKey",
    "GenericKey",
    "load_movielens",
    "load_movielens_df",
    "MTArray",
    "MTFloatArray",
    "MTGenericArray",
    "MTIntArray",
    "Vocabulary",
    "RecQuery",
    "QueryInput",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import ClassVar, Protocol

from lenskit.data import ItemList, ItemListCollection


class MetricFunction(Protocol):
    "Interface for per-list metrics implemented as simple functions."

    @abstractmethod
    def __call__(self, output: ItemList, test: ItemList, /) -> float: ...


class Metric(ABC):
    """
    Base class for LensKit metrics.  Individual metrics need to implement a
    sub-interface, such as :class:`ListMetric` and/or :class:`GlobalMetric`.

    For simplicity in the analysis code, you cannot simply implement the
    properties of this class on an arbitrary class in order to implement a
    metric with all available behavior such as labeling and defaults; you must
    actually extend this class.  This requirement may be relaxed in the future.

    Stability:
        Full
    """

    @property
    def label(self) -> str:
        """
        The metric's default label in output.

        The base implementation returns the class name by default.
        """
        return self.__class__.__name__

    def __str__(self):
        return f"Metric {self.label}"


class ListMetric(Metric):
    """
    Base class for metrics that measure individual recommendation (or
    prediction) lists, and whose results may be aggregated to compute overall
    metrics.

    For prediction metrics, this is *macro-averaging*.

    Stability:
        Full
    """

    default: ClassVar[float | None] = 0.0
    """
    The default value to infer when computing statistics over missing values.
    If ``None``, no inference is done (necessary for metrics like RMSE, where
    the missing value is theoretically infinite).
    """

    @abstractmethod
    def measure_list(self, output: ItemList, test: ItemList, /) -> float:
        """
        Compute the metric value for a single result list.

        Individual metric classes need to implement this method.
        """
        raise NotImplementedError()


class GlobalMetric(Metric):
    """
    Base class for metrics that measure entire runs at a time.

    For prediction metrics, this is *micro-averaging*.

    Stability:
        Full
    """

    @abstractmethod
    def measure_run(self, output: ItemListCollection, test: ItemListCollection, /) -> float:
        """
        Compute a metric value for an entire run.

        Individual metric classes need to implement this method.
        """
        raise NotImplementedError()


class DecomposedMetric(Metric):
    """
    Base class for metrics that measure entire runs through flexible
    aggregations of per-list intermediate measurements.  They can optionally
    extract individual-list metrics from the per-list measurements.

    Stability:
        Full
    """

    @abstractmethod
    def compute_list_data(self, output: ItemList, test: ItemList, /) -> object:
        """
        Compute measurements for a single list.
        """
        raise NotImplementedError()

    @abstractmethod
    def extract_list_metric(self, metric: object, /) -> float | None:
        """
        Extract a single-list metric from the per-list measurement result (if
        applicable).

        Returns:
            The per-list metric, or ``None`` if this metric does not compute
            per-list metrics.
        """
        return None

    @abstractmethod
    def global_aggregate(self, values: list[object], /) -> float:
        """
        Aggregate list metrics to compute a global value.
        """
        raise NotImplementedError()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Pydantic models for LensKit data schemas.  These models define define the data
schema in memory and also define how schemas are serialized to and from
configuration files.  See :ref:`data-model` for details.

.. note::

    The schema does not specify data types directly — data types are inferred
    from the underlying Arrow data structures.  This reduces duplication of type
    information and the opportunity for inconsistency.
"""

# pyright: strict
from __future__ import annotations

import re
from enum import Enum
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import BaseModel, StringConstraints, model_validator

NAME_PATTERN = re.compile(r"^[\w_]+$")
Name: TypeAlias = Annotated[str, StringConstraints(pattern=NAME_PATTERN)]


def id_col_name(name: str) -> str:
    return f"{name}_id"


def num_col_name(name: str) -> str:
    return f"{name}_num"


def check_name(name: str) -> None:
    """
    Check if a name is valid.

    Raises:
        ValueError: when the name is invalid.
    """
    if not NAME_PATTERN.match(name):
        raise ValueError(f"invalid name “{name}”")


class AllowableTroolean(Enum):
    """
    Three-way enumeration for storing both whether a feature is allowed and is
    used.  For convenience, in serialized data or configuration files these
    values may be specified either as strings or as booleans, in which case
    ``False`` is :attr:`FORBIDDEN` and ``True`` is :attr:`ALLOWED`.  They are
    always serialized as strings.
    """

    FORBIDDEN = "forbidden"
    """
    The feature is forbidden.
    """
    ALLOWED = "allowed"
    """
    The feature is allowed, but no records using it are present.
    """
    PRESENT = "present"
    """
    The feature is used by instances in the data.
    """

    @property
    def is_allowed(self) -> bool:
        """
        Query whether the feature is allowed.
        """
        return self != self.FORBIDDEN

    @property
    def is_forbidden(self) -> bool:
        """
        Query whether the feature is forbidden.
        """
        return self == self.FORBIDDEN

    @property
    def is_present(self) -> bool:
        """
        Query whether the feature is present (used in recorded instances).
        """
        return self == self.PRESENT

    @model_validator(mode="before")
    @classmethod
    def _validate_troolean(cls, value: Any):
        if value is True:
            return AllowableTroolean.ALLOWED
        elif value is False:
            return AllowableTroolean.FORBIDDEN
        else:
            return value


class AttrLayout(Enum):
    SCALAR = "scalar"
    """
    Scalar (non-list, non-vector) attribute value.
    """
    LIST = "list"
    """
    Homogenous, variable-length list of attribute values.
    """
    VECTOR = "vector"
    """
    Homogenous, fixed-length vector of numeric attribute values.
    """
    SPARSE = "sparse"
    """
    Homogenous, fixed-length sparse vector of numeric attribute values.
    """


class DataSchema(BaseModel, extra="forbid"):
    """
    Description of the entities and layout of a dataset.
    """

    name: str | None = None
    """
    The dataset name.
    """

    default_interaction: Name | None = None
    """
    The default interaction type.
    """

    entities: dict[Name, EntitySchema] = {}
    """
    Entity classes defined for this dataset.
    """
    relationships: dict[Name, RelationshipSchema] = {}
    """
    Relationship classes defined for this dataset.
    """


class EntitySchema(BaseModel, extra="forbid"):
    """
    Entity class definitions in the dataset schema.
    """

    id_type: Literal["int", "str"] | None = None
    """
    The data type for identifiers in this entity class.
    """
    attributes: dict[Name, ColumnSpec] = {}
    """
    Entity attribute definitions.
    """


class RelationshipSchema(BaseModel, extra="forbid"):
    """
    Relationship class definitions in the dataset schema.
    """

    entities: dict[Name, Name | None]
    """
    Define the entity classes participating in the relationship.  For aliased
    entity classes (necessary for self-relationships), the key is the alias, and
    the value is the original entity class name.
    """
    interaction: bool = False
    """
    Whether this relationship class records interactions.
    """
    repeats: AllowableTroolean = AllowableTroolean.FORBIDDEN
    """
    Whether this relationship supports repeated interactions.
    """
    attributes: dict[Name, ColumnSpec] = {}
    """
    Relationship attribute definitions.
    """

    @property
    def entity_class_names(self) -> list[str]:
        return list(self.entities.keys())


class ColumnSpec(BaseModel, extra="forbid"):
    layout: AttrLayout = AttrLayout.SCALAR
    """
    The attribute layout (whether and how multiple values are supported).
    """

    vector_size: int | None = None
    """
    The dimensionality of the vector, for sparse and vector columns.
    """



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import warnings
from typing import Any, Callable, Literal, overload

from ._base import Progress

_backend: Callable[..., Progress] = Progress


@overload
def set_progress_impl(name: Literal["rich"]): ...
@overload
def set_progress_impl(name: Literal["notebook"]): ...
def set_progress_impl(name: str | None, *options: Any):
    """
    Set the progress bar implementation.
    """
    global _backend

    match name:
        case "notebook":
            try:
                from ._notebook import JupyterProgress

                _backend = JupyterProgress
            except ImportError:
                warnings.warn("notebook progress backend needs ipywidgets")
                _backend = Progress

        case "rich":
            from ._rich import RichProgress

            _backend = RichProgress

        case "none" | None:
            _backend = Progress
        case _:
            raise ValueError(f"unknown progress backend {name}")


def item_progress(
    label: str, total: int | None = None, fields: dict[str, str | None] | None = None
) -> Progress:
    """
    Create a progress bar for distinct, counted items.

    Args:
        label:
            The progress bar label.
        total:
            The total number of items.
        fields:
            Additional fields to report with the progress bar (such as a current
            loss).  These are specified as a dictionary mapping field names to
            format strings (the pieces inside ``{...}`` in :meth:`str.format`),
            and the values come from extra kwargs to :meth:`Progress.update`;
            mapping to ``None`` use default ``str`` formatting.
    """
    return _backend(label, total, fields)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import click

from lenskit.data.dataset import Dataset
from lenskit.data.movielens import load_movielens
from lenskit.logging import get_logger

_log = get_logger(__name__)


@click.command("convert")
@click.option("--movielens", "format", flag_value="movielens", help="convert MovieLens data")
@click.argument("src", type=Path)
@click.argument("dst", type=Path)
def convert(format: str | None, src: Path, dst: Path):
    """
    Convert data into the LensKit native format.
    """

    match format:
        case None:
            _log.error("no data format specified")
            raise click.UsageError("no data format specified")
        case "movielens":
            data = convert_movielens(src)
        case _:
            raise ValueError(f"unknown data format {format}")

    log = _log.bind(dst=str(dst))
    log.info("saving data in native format")
    data.save(dst)


def convert_movielens(source: Path) -> Dataset:
    log = _log.bind(src=str(source))
    log.info("loading MovieLens data")
    return load_movielens(source)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2025 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
LensKit background monitoring.

Code almost never needs to interact directly with the monitoring subsystem.
"""

# pyright: basic
from __future__ import annotations

import atexit
import json
import logging
import multiprocessing as mp
import os.path
import pickle
import threading
import time
from contextlib import contextmanager
from enum import Enum
from hashlib import blake2b
from tempfile import TemporaryDirectory
from typing import Any, Protocol, runtime_checkable
from uuid import UUID, uuid4

import zmq

from ._proxy import get_logger
from .tasks import Task

SIGNAL_ADDR = "inproc://lenskit-monitor-signal"
REFRESH_INTERVAL = 5

_log = get_logger(__name__)
_monitor_lock = threading.Lock()
_monitor_instance: Monitor | None = None


@contextmanager
def maybe_close(sock: zmq.Socket[Any] | None):
    if sock is None:
        yield None
    else:
        try:
            yield sock
        finally:
            sock.close()


@runtime_checkable
class MonitorRefreshable(Protocol):
    def monitor_refresh(self):
        """
        Refresh this object in response to monitor refresh timeouts.
        """
        ...


def get_monitor() -> Monitor:
    """
    Get the monitor, starting it if it is not yet running.
    """
    from .worker import WorkerLogConfig

    global _monitor_instance

    if _monitor_instance is not None:
        return _monitor_instance

    with _monitor_lock:
        if _monitor_instance is None:
            ctx = WorkerLogConfig.current(from_monitor=False)
            _monitor_instance = Monitor(handle_logging=ctx is None)

        return _monitor_instance


@atexit.register
def _shutdown_monitor():  # type: ignore
    "Shut down the monitor, if one is running."
    global _monitor_instance
    if _monitor_instance is not None:
        _monitor_instance.shutdown()


class MonitorState(Enum):
    SHUTDOWN = 0
    ACTIVE = 1
    DRAINING = 2


class Monitor:
    """
    LensKit monitor controller.

    The monitor does several things:

    * Receive and re-inject log messages from worker processes.
    * Track work in progress and periodically write work logs.

    The monitor is managed and used internally, and neither LensKit client code
    nor component implementations often need to interact with it.

    Args:
        handle_logging:
            Whether or not to handle log messages.
    """

    zmq: zmq.Context[zmq.Socket[bytes]]
    _backend: MonitorThread
    _signal: zmq.Socket[bytes]
    log_address: str | None
    _tmpdir: TemporaryDirectory | None = None
    refreshables: dict[UUID, MonitorRefreshable]
    lock: threading.Lock

    def __init__(self, handle_logging: bool = True):
        self.zmq = zmq.Context()

        if handle_logging:
            addr, log_sock = self._log_sock()
            self.log_address = addr
        else:
            log_sock = None
            self.log_address = None

        self._backend = MonitorThread(self, log_sock)
        self._backend.start()

        self._signal = self.zmq.socket(zmq.REQ)
        self._signal.connect(SIGNAL_ADDR)

        self.lock = threading.Lock()
        self.refreshables = {}

        _log.bind(address=addr).info("monitor ready")

    def add_refreshable(self, obj: MonitorRefreshable) -> UUID:
        uuid = uuid4()
        with self.lock:
            self.refreshables[uuid] = obj
        return uuid

    def remove_refreshable(self, uuid: UUID):
        with self.lock:
            if uuid in self.refreshables:
                del self.refreshables[uuid]

    def shutdown(self):
        log = _log.bind()
        try:
            log.debug("requesting monitor shutdown")
            self._signal.send(b"shutdown")
            self._signal.close()

            log.debug("waiting for monitor thread to shut down")
            self._backend.join()

            log.info("monitor shut down")
            self.zmq.term()
        except Exception as e:
            self.zmq.destroy()
            raise e

    def _log_sock(self) -> tuple[str, zmq.Socket[bytes]]:
        sock = self.zmq.socket(zmq.PULL)
        if zmq.has("ipc"):
            self._tmpdir = TemporaryDirectory(prefix="lenskit-monitor.", ignore_cleanup_errors=True)
            path = os.path.join(self._tmpdir.name, "log-messages.sock")
            addr = f"ipc://{path}"
            sock.bind(addr)
        else:
            port = sock.bind_to_random_port("tcp://127.0.0.1")
            addr = f"tcp://127.0.0.1:{port}"
        log = _log.bind(address=addr)
        log.debug("bound listener socket")
        return addr, sock


class MonitorThread(threading.Thread):
    """
    LensKit monitoring backend thread.
    """

    monitor: Monitor
    state: MonitorState

    signal: zmq.Socket[bytes]
    log_sock: zmq.Socket[bytes] | None
    poller: zmq.Poller
    _authkey: bytes
    last_refresh: float

    def __init__(self, monitor: Monitor, log_sock: zmq.Socket[bytes] | None):
        super().__init__(name="LensKitMonitor", daemon=True)
        self.monitor = monitor
        self.log_sock = log_sock
        self._authkey = mp.current_process().authkey
        self.last_refresh = time.perf_counter()

    def run(self) -> None:
        self.state = MonitorState.ACTIVE
        _log.debug("monitor thread started")
        with maybe_close(self.log_sock), self.monitor.zmq.socket(zmq.PULL) as signal:
            signal.bind(SIGNAL_ADDR)
            self.signal = signal

            self.poller = zmq.Poller()
            self.poller.register(signal, zmq.POLLIN)
            if self.log_sock is not None:
                self.poller.register(self.log_sock, zmq.POLLIN)

            self._pump_messages()

            del self.poller

    def _pump_messages(self):
        timeout = 0
        last_refresh = time.perf_counter()

        while self.state != MonitorState.SHUTDOWN:
            # don't wait while draining
            if self.state == MonitorState.DRAINING:
                timeout = 0

            ready = dict(self.poller.poll(timeout))

            if self.signal in ready:
                self._handle_signal()

            if self.log_sock is not None and self.log_sock in ready:
                try:
                    self._handle_log_message()
                except Exception as e:
                    _log.error("error handling message", exc_info=e)

            if not ready:
                if self.state == MonitorState.DRAINING:
                    self.state = MonitorState.SHUTDOWN

                elif self.state == MonitorState.ACTIVE:
                    # nothing to do — check if we need a refresh
                    now = time.perf_counter()
                    left = max(int((REFRESH_INTERVAL - now + last_refresh) * 1000), 0)
                    if left < 20:
                        self._do_refresh()
                        timeout = REFRESH_INTERVAL * 1000
                    else:
                        timeout = left

    def _handle_signal(self):
        sig_msg = self.signal.recv()
        log = _log.bind(signal=sig_msg)
        match sig_msg:
            case b"":
                log.debug("stray empty message")
            case b"shutdown":
                log.debug("initiating shutdown")
                self.state = MonitorState.DRAINING
            case _:
                log.warning("unknown signal")

    def _handle_log_message(self):
        assert self.log_sock is not None
        parts = self.log_sock.recv_multipart()
        if len(parts) != 4:
            _log.warning("invalid multipart message, expected 3 parts")
            return
        engine, name, data, mac = parts

        mb = blake2b(key=self._authkey)
        mb.update(engine)
        mb.update(name)
        mb.update(data)
        if mb.digest() != mac:
            _log.warning("invalid log message digest, dropping")
            return

        engine = engine.decode()
        name = name.decode()

        if engine == "stdlib":
            rec = pickle.loads(data)
            logger = logging.getLogger(name)
            logger.handle(rec)
        elif engine == "structlog":
            logger = get_logger(name)
            data = json.loads(data)
            method = getattr(logger, data["method"])
            method(**data["event"])
        elif engine == "lenskit.logging.tasks":
            task = Task.model_validate_json(data)
            _log.debug("received subtask", task_id=str(task.task_id))
            current = Task.current()
            if current:
                current.add_subtask(task)
            else:
                _log.debug("no active task for subtask reporting")
        else:
            _log.error("invalid log backend")

    def _do_refresh(self):
        with self.monitor.lock:
            objs = list(self.monitor.refreshables.values())

        objs.reverse()

        for obj in objs:
            try:
                obj.monitor_refresh()
            except Exception as e:
                _log.warning("failed to refresh %s: %s", obj, e)
